/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* Author: Daniel Rypl */

#ifndef ENURBS_H
#define ENURBS_H

#ifdef __cplusplus
extern "C" {
#endif

#include <math.h>
#include <stdlib.h>
#include "Etypes.h"


#define MAX_NURBS_DEGREE     19     /* only for preallocation */


#define PRECISE_BBOX_NURBS          /* examines points on surface instead of ctrl points */

#ifdef PRECISE_BBOX_NURBS
	/* increase BB to account for inaccuracy due to examinig only some points on the curve/surface */
	//#define INCREASE_BBOX_NURBS         /* rather not recommended */
#define BBOX_PORTION_NURBS   20
#else
	/* increase BB to account for inaccuracy due to examining curve/surface with negative weight */
#define INCREASE_BBOX_NURBS         /* recommended */
#define BBOX_PORTION_NURBS   10
#endif

#ifdef INCREASE_BBOX_NURBS

/* causes that BB of surface reprezented by boundary curves only may be different from BB when also (or only) surface is handled;
	 this happens if BB of each boundary curve of the surface is smaller than BB of the surface itself */

#endif

#define GetFirstNonzeroBasisFunction(p,s) (s)-(p)
#define GetLastNonzeroBasisFunction(p,s) (s)

int 
GetSpan(int p, int n, double *U, double u);
void 
BasisFunctions(int p, int s, double *U, double u, double *bu);

#ifdef __cplusplus
}
#endif

#endif
