/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef ETYPEIN_H
#define ETYPEIN_H

#ifdef __cplusplus
extern "C" {
#endif


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "Etypes.h" /* viewer types */
#include "Ereprts.h"

/* choose appropriate version */

#define ELIXIR_TYPEIN_VERSION_BOOLEAN   /* original version - preferred because of compatibility with preceeding version */
	//#define ELIXIR_TYPEIN_VERSION_INT       /* compatibible with new Elixir using dynamic layers */



#if !defined(ELIXIR_TYPEIN_VERSION_BOOLEAN) && !defined(ELIXIR_TYPEIN_VERSION_INT)
#define ELIXIR_TYPEIN_VERSION_BOOLEAN   /* one ELIXIR_TYPEIN_VERSION must be defined */
#endif

#ifdef ELIXIR_TYPEIN_VERSION_BOOLEAN

#ifdef ELIXIR_TYPEIN_VERSION_INT
#undef ELIXIR_TYPEIN_VERSION_INT        /* only one ELIXIR_TYPEIN_VERSION must be defined */
#endif
/* original implementation:

	 BOOLEAN TypeInCompareTokenWithKeyword*(...);

	 avoid using statements as
	 if(TypeInCompareTokenWithKeyword*(...) == YES){...}
	 because they will work properly only in BOOLEAN version !!!
	 replace them by 
	 if(TypeInCompareTokenWithKeyword*(...) != NO){...}
	 which will work also in INT version as far as BOOLEAN as well as NO are defined as int
	 or by 
	 if(TypeInCompareTokenWithKeyword*(...)){...}
	 which will work in both versions

	 use of statement
	 if(TypeInCompareTokenWithKeyword*(...) == NO){...}
	 will work also in INT version as far as BOOLEAN as well as NO are defined as int
	 better alternative is
	 if(!TypeInCompareTokenWithKeyword*(...)){...}
	 which will work in both versions */
#endif

#ifdef ELIXIR_TYPEIN_VERSION_INT

#ifdef ELIXIR_TYPEIN_VERSION_BOOLEAN
#undef ELIXIR_TYPEIN_VERSION_BOOLEAN    /* only one ELIXIR_TYPEIN_VERSION must be defined */
#endif
/* implementation compatible with version of Elixir using dynamic number of layers:

	 int TypeInCompareTokenWithKeyword*(...);

	 avoid using statements as
	 if(TypeInCompareTokenWithKeyword*(...) == 5){...}
	 because they will work properly only in INT version !!!
	 replace them by 
	 if(TypeInCompareTokenWithKeyword*(...) != 0){...}
	 which will work also in BOOLEAN version as far as BOOLEAN is defined as int
	 or by 
	 if(TypeInCompareTokenWithKeyword*(...)){...}
	 which will work in both versions
	 
	 use of statement
	 if(TypeInCompareTokenWithKeyword*(...) == 0){...}
	 will work also in BOOLEAN version as far as BOOLEAN is defined as int
	 better alternative is
	 if(!TypeInCompareTokenWithKeyword*(...)){...}
	 which will work in both versions
*/
#endif

/* to avoid any ambiguities a set of specialized functions

	 #ifdef ELIXIR_TYPEIN_VERSION_INT
	 BOOLEAN TypeInCompareTokenWithKeyword*_BOOLEAN(...);
	 #endif

	 #ifdef ELIXIR_TYPEIN_VERSION_BOOLEAN
	 int TypeInCompareTokenWithKeyword*_INT(...);
	 #endif

	 is defined which can be together with the following redefinition 
	 (in appropriate application header)

	 #ifdef ELIXIR_TYPEIN_VERSION_INT
	 #define TypeInCompareTokenWithKeyword*_INT TypeInCompareTokenWithKeyword*
	 #endif

	 #ifdef ELIXIR_TYPEIN_VERSION_BOOLEAN
	 #define TypeInCompareTokenWithKeyword*_BOOLEAN TypeInCompareTokenWithKeyword*
	 #endif

	 safely used as

	 if(TypeInCompareTokenWithKeyword*_INT(...) == 5){...}
	 if(TypeInCompareTokenWithKeyword*_BOOLEAN(...) == YES){...}
*/

/* Elixir code is written in such a way that there is no need
	 to distinguished between BOOLEAN and INT version;
	 specialized functions are not used */


/* ===================================================================== */
/* Some convenience macros and global constants */

#ifdef MAX_LINE_LENGTH
#  undef MAX_LINE_LENGTH
#endif
#define MAX_LINE_LENGTH 256        /* maximum length of line in input file */

#ifdef SEPARATORS
#  undef SEPARATORS
#endif
#define SEPARATORS " "   /* separators of tokens */

#ifdef MAX_TOKENS
#  undef MAX_TOKENS
#endif
#define MAX_TOKENS 1000            /* maximum of tokens per line */

#ifdef COMMENT_CHAR
#  undef COMMENT_CHAR
#endif
#define COMMENT_CHAR '!'          /* comment char to be used in input file */

#define QUOTE '"'
#define BLANK ' '

/* ===================================================================== */

/* it would be nice if application can choose from following specifications and dynamically change it */

//#define TYPE_IN_NO_ERROR_REPORT
//#define TYPE_IN_NO_SYNTAX_HELP

#ifndef TYPE_IN_NO_ERROR_REPORT
//#define TYPE_IN_ERROR_REPORT_TO_STDERR
#define TYPE_IN_ERROR_REPORT_TO_WINDOW

#ifdef TYPE_IN_ERROR_REPORT_TO_WINDOW
#define TYPE_IN_ERROR_CLASS                   ELIXIR_ERROR_CLASS
#define TYPE_IN_ERROR_GRADE                   ERROR_GRADE
#endif
#endif

#ifndef TYPE_IN_NO_SYNTAX_HELP
//#define TYPE_IN_SYNTAX_HELP_TO_STDERR
#define TYPE_IN_SYNTAX_HELP_TO_WINDOW
#endif

typedef enum { NO_TOKEN, STRING, NUMBER } token_type;

typedef struct token_rec {
  int    token_type;
  char   *as_string;
  double  as_double;
} token_rec;

#define NO_TOKENS -1

typedef void (*exec_proc_p)(void);

typedef struct  _line_contents {
  int               tokens_total; /* >= 1         for non-empty command line */
				  /* == NO_TOKENS for empty line */
  int                line_number;
  token_rec tokens[MAX_TOKENS+1]; /* values of the tokens */
} line_contents;


typedef struct top_command {
  char              *keyword;
  int                keyword_length;
  exec_proc_p        handle_it;
  char              *syntax;
} top_command;
  
/* ===================================================================== */

void         TypeInPrintLineContents(void);
void         TypeInSyntaxHelp(int key);
int          TypeInBadToken( int ith, char *should_be_type );
void         TypeInBadSyntax( char *to_print );
void         TypeInBadParameter( char *to_print );
void         TypeInErrorInInput(void);
int          TypeInIsKeywordInLine( int start_at, char *keyword );
int          TypeInIsKeywordInLineMinLength( int start_at, char *keyword, int min_len );
#ifdef ELIXIR_INT_TYPEIN_VERSION
int          TypeInCompareTokenWithKeyword( char *tok, char *keyword);
int          TypeInCompareTokenWithKeywordMinLength(char *tok, char *keyword, int min_len);
int          TypeInCompareTokenWithKeywordExactMatch( char *tok, char *keyword);
/* specialized functions */
BOOLEAN      TypeInCompareTokenWithKeyword_BOOLEAN( char *tok, char *keyword);
BOOLEAN      TypeInCompareTokenWithKeywordMinLength_BOOLEAN(char *tok, char *keyword, int min_len);
BOOLEAN      TypeInCompareTokenWithKeywordExactMatch_BOOLEAN( char *tok, char *keyword);
#else
BOOLEAN      TypeInCompareTokenWithKeyword(char *tok, char *keyword);
BOOLEAN      TypeInCompareTokenWithKeywordMinLength(char *tok, char *keyword, int min_len);
BOOLEAN      TypeInCompareTokenWithKeywordExactMatch( char *tok, char *keyword);
/* specialized functions */
int          TypeInCompareTokenWithKeyword_INT( char *tok, char *keyword);
int          TypeInCompareTokenWithKeywordMinLength_INT(char *tok, char *keyword, int min_len);
int          TypeInCompareTokenWithKeywordExactMatch_INT( char *tok, char *keyword);
#endif
double       TypeInGetTokenAsDouble( int token_number );
int          TypeInGetTokenAsInt(int token_number);
long         TypeInGetTokenAsLong(int token_number);
char         *TypeInGetTokenAsString( int token_number );
token_type   TypeInGetTokenType( int token_number );
int          TypeInGetTokensTotal(void);
int          TypeInParseLine( char *buffer );
void         TypeInOpenInFile( char *in_file );
int          TypeInReadNextLine(void);
void         TypeInInitCommandLine(void);
int          TypeInCurrentCommand(void);
void         TypeInInputData( char *in_file );
void         TypeInInterpretCurrentCommand(void);
void         TypeInInterpretCurrentCommandEmpty(void);
top_command *TypeInAugmentCommandTable(top_command *table, int table_size);
top_command *TypeInReplaceCommand(top_command *cmd);
/* The table MUST start by { NULL, 0, TypeInErrorInInput }. */
void         TypeInSetCommandTable(top_command *table, int table_size);
top_command *TypeInGetCommandData(char *top_keyword);
int          TypeInCommandTableSize(void);

#ifdef __cplusplus
}
#endif

#endif
