/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* Author: Daniel Rypl */

#include "Econfig.h"
#include "Enurbs.h"


/* input:
	 p - degree
   n - number of control points
	 U - knot vector (of size m = n + p + 1)
	 u - parameter (U[0] <= u <= U[m-1])

	 output:
	 span (zero-based) s such that u is in <U[s];U[s+1])
	 except for s = pnts in which case u is in (U[s];U[s+1]> */

int 
GetSpan(int p, int n, double *U, double u)
{
	int low = p, high = n, mid;
  
  if(u == U[n])return(n - 1); 
  
  mid = (low + high) / 2;
  
  while(u < U[mid] || u >= U[mid+1]){
    if(u < U[mid])
      high = mid;
    else
			low = mid;
    mid = (low + high) / 2;
  }

  return(mid);
}

/* input:
	 p - degree
	 s - span (zero based)
	 U - knot vector (of size m = n + p + 1)
	 u - parameter (U[0] <= u <= U[m-1])
	 
	 output:
	 bu - array of p+1 values of nonzero basis functions */

#if 0
void BasisFunctions(int p, int s, double *U, double u, double *bu)
{
	int i, j;
	double right[MAX_NURBS_DEGREE + 1], *rr = right;
	double left[MAX_NURBS_DEGREE + 1], *ll = left;
	double saved, temp;

	if(p > MAX_NURBS_DEGREE){
		if((rr = calloc(p + 1, sizeof(double))) == NULL){
			fprintf(stderr, "Memory allocation failed in BasisFunctions\n");
			exit(0);
		}
		if((ll = calloc(p + 1, sizeof(double))) == NULL){
			fprintf(stderr, "Memory allocation failed in BasisFunctions\n");
			exit(0);
		}
	}

  bu[0] = 1.0;
  for(i = 1; i <= p; i++){
    left[i] = u - U[s + 1 - i];
    right[i] = U[s + i] - u;
    saved = 0.0;
    for(j = 0; j < i; j++) {
      temp = bu[j] / (right[j + 1] + left[i - j]);
      bu[j] = saved + right[j + 1] * temp;
      saved = left[i - j] * temp;
    }
    bu[i] = saved;
  }

	if(p > MAX_NURBS_DEGREE){
		free(rr);
		free(ll);
	}
}
#endif

#if 1
void BasisFunctions(int p, int s, double *U, double u, double *bu)
{
	int i, j;
	double right[MAX_NURBS_DEGREE], *rr = right;
	double left[MAX_NURBS_DEGREE], *ll = left;
	double saved, temp;

	if(p > MAX_NURBS_DEGREE){
		if((rr = calloc(p, sizeof(double))) == NULL){
			fprintf(stderr, "Memory allocation failed in BasisFunctions\n");
			exit(0);
		}
		if((ll = calloc(p, sizeof(double))) == NULL){
			fprintf(stderr, "Memory allocation failed in BasisFunctions\n");
			exit(0);
		}
	}

  bu[0] = 1.0;
  for(i = 0; i < p; i++){
    left[i] = u - U[s - i];
    right[i] = U[s + i + 1] - u;
    saved = 0.0;
    for(j = 0; j < i + 1; j++) {
      temp = bu[j] / (right[j] + left[i - j]);
      bu[j] = saved + right[j] * temp;
      saved = left[i - j] * temp;
    }
    bu[i + 1] = saved;
  }

	if(p > MAX_NURBS_DEGREE){
		free(rr);
		free(ll);
	}
}
#endif
