/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "Econfig.h"  
#include <ctype.h>
#include "Etypein.h"
#include "tokensP.h" /* of Ckit library */
#include "Esimple.h"

/* ========================================================================= */
/* PRIVATE */
/* ========================================================================= */

static top_command *top_command_table = NULL;

static int commands_total = 0;	/* number of fields in the commands[] array */

 /* FILE pointer for input */
static FILE *in_FILE_p = NULL;


static line_contents command_line; /* command-line data structure */


/* ========================================================================= */
/* PUBLIC */
/* ========================================================================= */

/* ========================================================================= */

void 
TypeInInputData(char *in_file) /* TypeInSetCommandTable must be */
				      /* called as the first thing before */
				      /* entering TypeInInputData */
{
  TypeInInitCommandLine();
  TypeInOpenInFile(in_file);
  while(TypeInReadNextLine() != EOF) {
    top_command_table[TypeInCurrentCommand()].handle_it();
  }
}

/* ========================================================================= */

void 
TypeInInterpretCurrentCommand()
{
  /* TypeInPrintLineContents(); */ /* uncomment for debugging purposes */
  top_command_table[TypeInCurrentCommand()].handle_it();
}


void
handle_command(Widget w, char *s);


/* added by dr to echo processed active set commands */

void 
TypeInInterpretCurrentCommandEmpty()
{
	int tokens = command_line.tokens_total;
#if 0
	char buffer[64];

	sprintf(buffer, "%s", top_command_table[TypeInCurrentCommand()].keyword);
	handle_command(ESITopLevelWidget(), buffer);
#endif
	command_line.tokens_total = 1;
	TypeInInterpretCurrentCommand();
	command_line.tokens_total = tokens;
}
 
/* ========================================================================= */
/* The table MUST start by { NULL, 0, TypeInErrorInInput }. */

void 
TypeInSetCommandTable(top_command *table, int table_size)
{
  top_command_table = table;
  commands_total    = table_size;
  TypeInInitCommandLine();
}

/* ========================================================================= */
/* Works similarly to TypeInSetCommandTable when there was no command */
/* table active; otherwise appends the new command table at the end. */
/* The table must not start by { NULL, 0, TypeInErrorInInput }. */

top_command *
TypeInAugmentCommandTable(top_command *table, int table_size)
{
  top_command *old_table, *new_table;
  int old_commands_total, new_commands_total, i, j;

  if (table == NULL)
    return top_command_table;   /* request to get the current command table */
  
  old_table          = top_command_table;
  old_commands_total = commands_total;
  if (old_table == NULL) {
    old_commands_total = 1; /* take account of the nil command */
  }
  new_commands_total = old_commands_total + table_size;

  new_table = (top_command *)make_node(sizeof(top_command)*new_commands_total);
  if (new_table == NULL) {
    fprintf(stderr, "Failed allocate memory\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  if (old_table == NULL) {
    new_table[0].keyword        = NULL;
    new_table[0].keyword_length = 0;
    new_table[0].handle_it      = TypeInErrorInInput;
    new_table[0].syntax         = NULL;
  } else {
    for (i = 0; i < old_commands_total; i++) {
      if (old_table[i].keyword != NULL)
        new_table[i].keyword        = strdup(old_table[i].keyword);
      else
        new_table[i].keyword        = NULL;
      new_table[i].keyword_length = old_table[i].keyword_length;
      new_table[i].handle_it      = old_table[i].handle_it;
      if (old_table[i].syntax != NULL)
        new_table[i].syntax         = strdup(old_table[i].syntax);
      else
        new_table[i].syntax         = NULL;
    }
  }
  for (j = 0; j < table_size; j++) {
    i = j + old_commands_total;
    new_table[i].keyword        = strdup(table[j].keyword);
    new_table[i].keyword_length = table[j].keyword_length;
    new_table[i].handle_it      = table[j].handle_it;
    if (table[j].syntax != NULL)
      new_table[i].syntax         = strdup(table[j].syntax);
    else
      new_table[i].syntax         = NULL;
  }
  top_command_table = new_table;
  commands_total    = new_commands_total;
  TypeInInitCommandLine();
  return old_table;
}

/* ========================================================================= */

/* modified by dr to identify ambiguous command */

int 
TypeInCurrentCommand()
{
  int i;
  char *tok;
  int chosen = 0;
	int tok_len;

  tok = TypeInGetTokenAsString(1);
	tok_len = strlen(tok);

  /* Note that we start at i = 1 (the error_handler is at the zero index) */
  for(i = 1; i < commands_total; i++) {
    if (TypeInCompareTokenWithKeyword (tok, top_command_table[i].keyword)){
			if (tok_len == strlen(top_command_table[i].keyword))return i;   /* exact match */
			if (chosen != 0)return 0;   /* ambiguous keyword (envokes error_handler) */
			chosen = i;
    }
  }

  return chosen;  /* note: zero envokes error_handler */
}

/* ========================================================================= */

void 
TypeInInitCommandLine()
{
  int i;

  if (top_command_table == NULL) {
    fprintf(stderr, "No command table!?\n");
    exit (ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  for (i = 0; i <= MAX_TOKENS; i++) {
    command_line.line_number          = 0;
    command_line.tokens[i].token_type = NO_TOKEN;
    command_line.tokens[i].as_string  = NULL;
    command_line.tokens[i].as_double  = 0.0;
  }
}

/* ========================================================================= */

int 
TypeInReadNextLine() /* returns YES for valid input obtained or EOF */
{
  char buffer[MAX_LINE_LENGTH];
  int anything_to_return = NO;

  while(anything_to_return == NO) {
    command_line.line_number++;
    if (fgets(buffer, MAX_LINE_LENGTH-1, in_FILE_p) == NULL)	{
      if (feof(in_FILE_p))
	return EOF;
    }
    anything_to_return = TypeInParseLine(buffer);
  }
  return YES;
}

/* ========================================================================= */

void 
TypeInOpenInFile(char *in_file) 
{
  if ((in_FILE_p = fopen(in_file, "r")) == NULL) {
    fprintf(stderr, "\nFailed to open input file \"%s\".", in_file);
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
}

/* ========================================================================= */

int 
TypeInParseLine(char *to_be_parsed)
{
  char           tok[MAX_LINE_LENGTH];
  int have_input = NO;
  int             ptr = 0;
  char           *chp;
  double      num_val;
  int          tok_ix, tok_type;

	command_line.tokens_total = 0;	
  if (to_be_parsed == NULL) return NO;
  
  tok_ix      = NO_TOKENS;
  while ((tok_type = tokens_get_token(to_be_parsed, &ptr, tok, SKIP_SPACES))
         != ENDOFLINE_TOKEN) {
    if (tok_type == NO_TOKEN_AT_ALL)
      break;
    
    tok_ix++;
    if ((command_line.tokens[tok_ix].as_string) != NULL)	{
      free(command_line.tokens[tok_ix].as_string);
      command_line.tokens[tok_ix].as_string = NULL;
    }
    if (tok[0] == COMMENT_CHAR)
      break;  /* this token is comment char -- return  */
    /* Assume string */
    command_line.tokens[tok_ix].as_string  = strdup(tok);
    command_line.tokens[tok_ix].token_type = STRING;
    if (tok_type == NUMBER_TOKEN) {
      errno = 0;
      num_val = strtod(tok, &chp);
      if (chp != tok && errno == 0) { /* successful conversion */
        command_line.tokens[tok_ix].as_double  = num_val;
        command_line.tokens[tok_ix].token_type = NUMBER;
      }
    }
    command_line.tokens_total = tok_ix + 1;
    have_input = YES;
  }
  return have_input;
}

/* ========================================================================= */

int 
TypeInGetTokensTotal()
{
  return command_line.tokens_total;
}

/* ========================================================================= */

token_type 
TypeInGetTokenType(int token_number)
{
  if (token_number <= command_line.tokens_total)
    return command_line.tokens[token_number-1].token_type;
  else
    return NO_TOKEN;
}

/* ========================================================================= */

char *
TypeInGetTokenAsString(int token_number)
{
  if (token_number <= command_line.tokens_total)
    return command_line.tokens[token_number-1].as_string;
  else
    return NULL;
}

/* ========================================================================= */

double 
TypeInGetTokenAsDouble(int token_number)
{
  if (token_number <= command_line.tokens_total)  {
    if (command_line.tokens[token_number-1].token_type == NUMBER)
      return command_line.tokens[token_number-1].as_double;
  }
  return 0.0;
}     

/* ========================================================================= */

int 
TypeInGetTokenAsInt(int token_number)
{
  if (token_number <= command_line.tokens_total){
    if (command_line.tokens[token_number-1].token_type == NUMBER)
      return (int)(command_line.tokens[token_number-1].as_double);
  }
  return 0;
}     

/* added by dr */

long
TypeInGetTokenAsLong(int token_number)
{
  if (token_number <= command_line.tokens_total){
    if (command_line.tokens[token_number-1].token_type == NUMBER)
      return (long)(command_line.tokens[token_number-1].as_double);
  }
  return 0;
}     

/* ========================================================================= */

void
TypeInSyntaxHelp(int key)
{
	char buffer[64];

/* do not use TypeInGetTokenAsString(1) here if syntax help is printed multiple times;
	 after first use, the next one(s) would print help on help or guide */

	if(key == 0){
#ifdef TYPE_IN_SYNTAX_HELP_TO_WINDOW
		ESIHandleCmd("help");
#endif
#ifdef TYPE_IN_SYNTAX_HELP_TO_STDERR
		ESIHandleCmd("guide");
#endif
	}
	else{
#ifdef TYPE_IN_SYNTAX_HELP_TO_WINDOW
		sprintf(buffer, "help %s", top_command_table[key].keyword);
		ESIHandleCmd(buffer);
#endif
#ifdef TYPE_IN_SYNTAX_HELP_TO_STDERR
		sprintf(buffer, "guide %s", top_command_table[key].keyword);
		ESIHandleCmd(buffer);
#endif
	}
}


/* ========================================================================= */

int
TypeInBadToken(int ith, char *should_be_type)
{
	int key;
	char buffer[256];

	key = TypeInCurrentCommand();

#ifdef TYPE_IN_ERROR_REPORT_TO_WINDOW

	sprintf(buffer, "Bad token type of command %s in line #%d\n  %dth token = \"%s\" (should be a %s)\n",
					/* TypeInGetTokenAsString(1), */
					top_command_table[key].keyword, 
					command_line.line_number, ith, TypeInGetTokenAsString(ith), should_be_type);
	ERptErrMessage(TYPE_IN_ERROR_CLASS, 1, buffer, TYPE_IN_ERROR_GRADE);
#endif
#ifdef TYPE_IN_ERROR_REPORT_TO_STDERR
  fprintf(stderr, "\nBad token type of command %s in line #%d\n  %dth token = \"%s\" (should be a %s)\n",
					/* TypeInGetTokenAsString(1), */
					top_command_table[key].keyword, 
					command_line.line_number, ith, TypeInGetTokenAsString(ith), should_be_type);
#endif

	TypeInSyntaxHelp(key);
  return 0;
}

/* ========================================================================= */

void 
TypeInBadSyntax(char *to_print)
{
	int key;
	char buffer[512];

	key = TypeInCurrentCommand();

#ifdef TYPE_IN_ERROR_REPORT_TO_WINDOW
  sprintf(buffer, "Bad or ambiguous syntax of command %s in line #%d\n",
					/* TypeInGetTokenAsString(1), */
					top_command_table[key].keyword, 
					command_line.line_number);
	if(to_print != NULL)strcat(buffer, to_print);   /* no check on buffer overflow !!! */
	ERptErrMessage(TYPE_IN_ERROR_CLASS, 1, buffer, TYPE_IN_ERROR_GRADE);
#endif
#ifdef TYPE_IN_ERROR_REPORT_TO_STDERR
  fprintf(stderr, "\nBad or ambiguous syntax of command %s in line #%d\n",
					/* TypeInGetTokenAsString(1), */
					top_command_table[key].keyword, 
					command_line.line_number);
	if(to_print != NULL)fprintf(stderr, "  %s\n", to_print);
#endif

	TypeInSyntaxHelp(key);
}

/* ========================================================================= */

void 
TypeInBadParameter(char *to_print)
{
	int key;
	char buffer[512];

	key = TypeInCurrentCommand();

#ifdef TYPE_IN_ERROR_REPORT_TO_WINDOW
  sprintf(buffer, "Bad parameter of command %s in line #%d\n",
					/* TypeInGetTokenAsString(1), */
					top_command_table[key].keyword, 
					command_line.line_number);
	if(to_print != NULL)strcat(buffer, to_print);   /* no check on buffer overflow !!! */
	ERptErrMessage(TYPE_IN_ERROR_CLASS, 1, buffer, TYPE_IN_ERROR_GRADE);
#endif
#ifdef TYPE_IN_ERROR_REPORT_TO_STDERR
  fprintf(stderr, "\nBad parameter of command %s in line #%d\n",
					/* TypeInGetTokenAsString(1), */
					top_command_table[key].keyword, 
					command_line.line_number);
	if(to_print != NULL)fprintf(stderr, "  %s\n", to_print);
#endif

	TypeInSyntaxHelp(key);
}

/* ========================================================================= */

void
TypeInErrorInInput()
{
#ifdef TYPE_IN_ERROR_REPORT_TO_WINDOW
	char buffer[256];

  sprintf(buffer, "Unknown or ambiguous top keyword \"%s\" in line #%d. Ignoring.\n",
					TypeInGetTokenAsString(1), command_line.line_number);
	ERptErrMessage(TYPE_IN_ERROR_CLASS, 1, buffer, TYPE_IN_ERROR_GRADE);
#endif
#ifdef TYPE_IN_ERROR_REPORT_TO_STDERR
  fprintf(stderr, "\nUnknown or ambiguous top keyword \"%s\" in line #%d. Ignoring.\n",
	  TypeInGetTokenAsString(1), command_line.line_number);
#endif

	TypeInSyntaxHelp(0);
}

/* ========================================================================= */

void 
TypeInPrintLineContents()
{
  int tot, i;
  token_type type;
  
  printf("\nLine # %d --> number of tokens: %d",
	 command_line.line_number, tot = TypeInGetTokensTotal());
  for (i=1; i <= tot; i++){
    type = TypeInGetTokenType(i);
    switch((int)type)	{
    case (int)NUMBER:
      printf("\n  %d type = NUMBER, val = %f",
	     i,TypeInGetTokenAsDouble(i));
      break;
    case (int)STRING:
      printf("\n  %d type = STRING, val = %s",
	     i,TypeInGetTokenAsString(i));
      break;
    default:
      printf("\n  %d type=UNKNOWN",i);
    }
  }
  printf("\n==========================================\n");
}

/* ========================================================================= */

/*
   Keyword      Token      Match?
    MV           MVNUMS     no
    MVNUM        MVNUMS     no (note that the S is missing in keyword)
    MV           MV         yes
    MVNUMS       MV         yes, but provided token `MVNUMS' 
                            is also present it will be chosen 
                            as a better match
    MVNUMS       MVN        yes (remark above applies)
 */

/* if there are two or more identical tokens that match the keyword, the first relevant is returned;
	 eg. "view rotnor z 0.1 z" returns 3 when calling TypeInIsKeywordInLine(2, "Z");
	 and returns 5 when calling TypeInIsKeywordInLine(4, "Z"); */

int 
TypeInIsKeywordInLine(int start_at, char *keyword)
{
  int i, tot;
  char  *tok;
  int chosen = 0;
  int matched_for_chosen = 0;
  int key_len = strlen(keyword);
  int tok_len;
 
  for(i = (start_at > 0)?start_at:1, tot = TypeInGetTokensTotal(); i <= tot; i++)  {
    tok = TypeInGetTokenAsString(i);
    tok_len = strlen(tok);
		if (tok_len > matched_for_chosen){
			if (TypeInCompareTokenWithKeyword (tok, keyword)){
				if (tok_len == key_len)return i;      /* exact match */
				if (tok_len > matched_for_chosen){    /* better match */
					chosen = i;
					matched_for_chosen = tok_len;
				}
			}
		}
	}

  return chosen;
}

/* ========================================================================= */

/* added by dr */

int 
TypeInIsKeywordInLineMinLength(int start_at, char *keyword, int min_len)
{
  int i, tot;
  char  *tok;
  int chosen = 0;
  int matched_for_chosen = 0;
  int key_len = strlen(keyword);
  int tok_len;
 
  for(i = (start_at > 0)?start_at:1, tot = TypeInGetTokensTotal(); i <= tot; i++)  {
    tok = TypeInGetTokenAsString(i);
    tok_len = strlen(tok);
		if (tok_len > matched_for_chosen && tok_len >= min_len){
			if (TypeInCompareTokenWithKeyword (tok, keyword)){
				if (tok_len == key_len)return i;      /* exact match */
				chosen = i;                           /* better match */
				matched_for_chosen = tok_len;
			}
		}
	}

  return chosen;
}

/* ========================================================================= */

/* modified by dr;
	 returns 0 or tok length;
	 nonzero cannot be returned if keyword is shorter than tok */

int
#ifdef ELIXIR_TYPEIN_VERSION_INT
TypeInCompareTokenWithKeyword(char *tok, char *keyword)
#endif
#ifdef ELIXIR_TYPEIN_VERSION_BOOLEAN
TypeInCompareTokenWithKeyword_INT(char *tok, char *keyword)
#endif
{
  int j, len;
  
	len = strlen(tok);
	if(len > strlen(keyword))return 0;

  for(j = 0; j < len; j++) {
    if (toupper (*(tok+j)) != toupper (*(keyword+j)))	{
      return 0;
    }
  }
  return len;
}

/* ========================================================================= */

/* modified by dr;
	 returns 0 or tok length;
	 nonzero cannot be returned if keyword is shorter than tok or if min_len is "longer" than tok */

/* note: statements
	   if(TypeInCompareTokenWithKeywordMinLength(tok, keyword, min_len) != 0)
	 and 
     if(TypeInCompareTokenWithKeyword(tok, keyword) > min_len)
	 are equivalent;
	 call to TypeInCompareTokenWithKeywordMinLength is more efficient
	 because the comparison is not performed if tok is too short */

int
#ifdef ELIXIR_TYPEIN_VERSION_INT
TypeInCompareTokenWithKeywordMinLength(char *tok, char *keyword, int min_len)
#endif
#ifdef ELIXIR_TYPEIN_VERSION_BOOLEAN
TypeInCompareTokenWithKeywordMinLength_INT(char *tok, char *keyword, int min_len)
#endif
{
  int j, len;
  
	len = strlen(tok);
	if(len < min_len || len > strlen(keyword))return 0;

  for(j = 0; j < len; j++) {
    if (toupper (*(tok+j)) != toupper (*(keyword+j)))	{
      return 0;
    }
  }
  return len;
}

/* ========================================================================= */

/* returns 0 or tok length = keyword length */

int
#ifdef ELIXIR_TYPEIN_VERSION_INT
TypeInCompareTokenWithKeywordExactMatch(char *tok, char *keyword)
#endif
#ifdef ELIXIR_TYPEIN_VERSION_BOOLEAN
TypeInCompareTokenWithKeywordExactMatch_INT(char *tok, char *keyword)
#endif
{
  int j, len;
  
	len = strlen(tok);
	if(len != strlen(keyword))return 0;

  for(j = 0; j < len; j++) {
    if (toupper(*(tok+j)) != toupper (*(keyword+j)))	{
      return 0;
    }
  }
  return len;
}


/* modified by dr;
	 YES cannot be returned if keyword is shorter than tok */

BOOLEAN 
#ifdef ELIXIR_TYPEIN_VERSION_BOOLEAN
TypeInCompareTokenWithKeyword(char *tok, char *keyword)
#endif
#ifdef ELIXIR_TYPEIN_VERSION_INT
TypeInCompareTokenWithKeyword_BOOLEAN(char *tok, char *keyword)
#endif
{
  int j, len;
  
	len = strlen(tok);
	if(len > strlen(keyword))return NO;

  for(j = 0; j < len; j++) {
    if (toupper(*(tok+j)) != *(keyword+j))	{
      return NO;
    }
  }
  return YES;
}

/* ========================================================================= */

/* added by dr
	 YES cannot be returned if keyword is shorter than tok or if min_len is "longer" than tok  */

BOOLEAN
#ifdef ELIXIR_TYPEIN_VERSION_BOOLEAN
TypeInCompareTokenWithKeywordMinLength(char *tok, char *keyword, int min_len)
#endif
#ifdef ELIXIR_TYPEIN_VERSION_INT
TypeInCompareTokenWithKeywordMinLength_BOOLEAN(char *tok, char *keyword, int min_len)
#endif
{
  int j, len;
  
	len = strlen(tok);
	if(len < min_len || len > strlen(keyword))return NO;

  for(j = 0; j < len; j++) {
    if (toupper (*(tok+j)) != toupper (*(keyword+j)))	{
      return 0;
    }
  }
  return YES;
}

/* ========================================================================= */

/* added by dr */

BOOLEAN
#ifdef ELIXIR_TYPEIN_VERSION_BOOLEAN
TypeInCompareTokenWithKeywordExactMatch(char *tok, char *keyword)
#endif
#ifdef ELIXIR_TYPEIN_VERSION_INT
TypeInCompareTokenWithKeywordExactMatch_BOOLEAN(char *tok, char *keyword)
#endif
{
  int j, len;
  
	len = strlen(tok);
	if(len != strlen(keyword))return NO;

  for(j = 0; j < len; j++) {
    if (toupper(*(tok+j)) != toupper (*(keyword+j)))	{
      return NO;
    }
  }
  return YES;
}

/* ========================================================================= */

/* modified by dr to identify ambiguous command */

top_command *
TypeInGetCommandData(char *top_keyword)
{
  int i;
	int chosen = 0;
	int key_len = strlen(top_keyword);

  /* Note that we start at i = 1 (the error_handler is at the zero index) */
  for (i = 1; i < commands_total; i++){
		if (TypeInCompareTokenWithKeyword(top_keyword, top_command_table[i].keyword)){
			if (key_len == strlen(top_command_table[i].keyword))return &top_command_table[i]; /* exact match */
			if (chosen != 0)return NULL;  /* ambiguous keyword */
			chosen = i;
		}
	}
  if(chosen == 0)return NULL; /* No command table or no such keyword */

	return &top_command_table[chosen];
}

/* modified by dr to accept only exact match */

top_command *
TypeInReplaceCommand(top_command *cmd)
{
  int i;
  
  if (top_command_table != NULL) {
    for (i = 1; i < commands_total; i++)
      if (TypeInCompareTokenWithKeywordExactMatch(cmd->keyword, top_command_table[i].keyword)) {
        /* The command was already in the table */
        top_command_table[i].handle_it = cmd->handle_it;
        if (top_command_table[i].syntax != NULL)
          free(top_command_table[i].syntax);
        if (cmd->syntax != NULL)
          top_command_table[i].syntax = strdup(cmd->syntax);
        return NULL;
      }
  } 
  /* The command is new */
  return TypeInAugmentCommandTable(cmd, 1);
}


int 
TypeInCommandTableSize(void)
{
  return commands_total;
}
